// texture.cpp
// by Glenn G. Chappell
// November 2003
//
// For CS 381
// Example of Texture Mapping

#include <iostream>
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <stdlib.h>  // Some versions of MS-Vis C++ have broken <cstdlib>
//using std::exit;
#include <GL/glut.h> // GLUT stuff, includes OpenGL headers as well


// Global variables
// Window/viewport
const int startwinsize = 400; // Starting window width & height, in pixels
const double neardist = 0.5;  // Near & far clippng planes
const double fardist = 10.;

// Keyboard
const int ESCKEY = 27;        // ASCII value of escape character

// For texture
const int img_width = 32;
const int img_height = 32;
GLubyte the_image[img_height][img_width][3];
   // The texture image
   // 3rd subscript 0 = R, 1 = G, 2 = B
GLuint texname;  // Name for this texture

// Lighting
bool lighton = true;   // True if we are lit
GLfloat light0_maincolor[] = { 1., 1., 1., 1. };
GLfloat light0_ambcolor[] = { 0., 0., 0., 0. };
GLfloat light0_position[] = { -2., 2., 3., 1. };
   // Above are parameters for GL_LIGHT0

// For other user-controlled stuff
bool textureon = true; // True if we are texturing

double zoom = 1.0;    // scale factor
const double minzoom = 0.1;
const double maxzoom = 10.0;
const double zoomfactor = 1.1; // Mult. factor for zoom increase

bool rotate = false;  // Is rotating?
double angle = 0.0;   // Current rotation angle (degrees)
const double anglestep = 1.0;  // Step for above (degrees)


// setmaterial_main
// Sets up "main" material
void setmaterial_main()
{
   GLfloat paintcolor[] = { 0.9, 0.9, 0.9, 1.0 };
   GLfloat specularcolor[] = { 1.0, 1.0, 1.0, 1.0 };
   glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, paintcolor);
   glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specularcolor);
   glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128.);
      // Set the shininess to 1 to find out why we like raising the
      // cosine to a high power.
}
// TRY THIS: Change the paint color above, and see how it affects the
// appearance of the texture.


// printbitmap
// Prints the given string at the given raster position
//  using GLUT bitmap fonts.
// You probably don't want any rotations in the model/view
//  transformation when calling this function.
void printbitmap(const string msg, double x, double y)
{
   glRasterPos2d(x, y);
   for (string::const_iterator ii = msg.begin();
        ii != msg.end();
        ++ii)
   {
      glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *ii);
   }
}


// display
// The GLUT display function
void display()
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glLightfv(GL_LIGHT0, GL_POSITION, light0_position);  // Stationary light

   // Draw object
   glPushMatrix();  // We are in model/view mode

      glTranslated(0., 0., -2.);

      //glMatrixMode(GL_TEXTURE);
      //glLoadIdentity();
      //glRotated(angle, 0,0,1);
      // TRY THIS: Comment out the glRotate* line below,
      //  and uncomment the above 3 lines.
      glRotated(angle, 1,1,0);
      glScaled(zoom, zoom, zoom);

      glMatrixMode(GL_MODELVIEW);
         // The above is not normally needed.
         // But it doesn't hurt, and it IS needed if you try
         // making the above-suggested change.

      if (lighton)
         glEnable(GL_LIGHTING);
      else
         glDisable(GL_LIGHTING);

      if (textureon)
         glEnable(GL_TEXTURE_2D);
      else
         glDisable(GL_TEXTURE_2D);

      glColor3d(0.9, 0.5, 0.1);  // Not used when lighting is on
      setmaterial_main();

      glutSolidTeapot(0.5);

   glPopMatrix();

   // Draw instructions
   glColor3d(0.0, 0.0, 0.0);

   glDisable(GL_TEXTURE_2D);  // Don't want to affect colors of remaining stuff
   glDisable(GL_DEPTH_TEST);
   glDisable(GL_LIGHTING);
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();              // Save old projection matrix
      glLoadIdentity();
      gluOrtho2D(-1., 1., -1., 1.);  // old faithful
      glColor3d(0., 0., 0.);
      printbitmap("Texture Demo", -0.9, 0.9);
      printbitmap("R    Toggle rotation", -0.9, 0.8);
      printbitmap("T    Toggle texture", -0.9, 0.7);
      printbitmap("L    Toggle lighting", -0.9, 0.6);
      printbitmap("Esc  Quit", -0.9, 0.5);
   glPopMatrix();               // Restore old projection matrix
   glMatrixMode(GL_MODELVIEW);  // Always go back to model/view mode
   glEnable(GL_DEPTH_TEST);     // Back to "normal" for 3-D drawing
   // Lighting, texture enabling are done at beginning of display

   glutSwapBuffers();
}


// reshape
// The GLUT reshape function
void reshape(int w, int h)
{
    glViewport(0, 0, w, h);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(60, double(w)/h, neardist, fardist);

   glMatrixMode(GL_MODELVIEW);  // Always go back to modelview mode
}


// keyboard
// The GLUT keyboard function
void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESCKEY:  // ESC: Quit
      exit(0);
      break;
   case 'r':     // R: Toggle rotation
   case 'R':
      rotate = !rotate;
      glutPostRedisplay();
      break;
   case 't':     // T: Toggle texture
   case 'T':
      textureon = !textureon;
      glutPostRedisplay();
      break;
   case 'l':     // L: Toggle lighting
   case 'L':
      lighton = !lighton;
      glutPostRedisplay();
      break;
   }
}


// special
// The GLUT special function
void special(int key, int x, int y)
{
   switch(key)
   {
   case GLUT_KEY_RIGHT: // Zoom goes up
      zoom *= zoomfactor;
      if (zoom > maxzoom) zoom = maxzoom;
      glutPostRedisplay();
      break;
   case GLUT_KEY_LEFT:  // Zoom goes down
      zoom /= zoomfactor;
      if (zoom < minzoom) zoom = minzoom;
      glutPostRedisplay();
      break;
   }
}


// idle
// The GLUT idle function
void idle()
{
   // Print OpenGL errors, if there are any (for debugging)
   if (GLenum err = glGetError())
   {
      cerr << "OpenGL ERROR: " << gluErrorString(err) << endl;
   }

   // Do rotation
   if (rotate)
   {
      angle += anglestep;
      glutPostRedisplay();
   }
}


// makeimage
// Make image in the_image
// (Code taken without change from drawimage.cpp)
void makeimage()
{
   for (int i=0; i<img_width; ++i)
   {
      double x = double(i)/(img_width-1);
      for (int j=0; j<img_height; ++j)
      {
         double y = double(j)/(img_height-1);

         the_image[j][i][0] = int(x*255)*15 % 256;
         the_image[j][i][1] = int(y*255)*15 % 256;
         the_image[j][i][2] = 0.75*255;
      }
   }
}


// init
// Initialization
// Called by main after window creation
void init()
{
   glClearColor(1.0, 1.0, 1.0, 0.0);
   glEnable(GL_DEPTH_TEST);

   // Lighting
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_maincolor);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light0_maincolor);
   glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambcolor);
      // Light 0 position is set in display
   glEnable(GL_LIGHT0);
   glEnable(GL_LIGHTING);
   // Texture set-up

   // Basic image set-up
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
   makeimage();

   // Generate a texture name and bind it to GL_TEXTURE_2D.
   // Then we no longer use texname, since GL_TEXTURE_2D is bound to it.
   // NOTE: binding a texture name is actually NOT necessary when we
   //  use only one texture, but it does not hurt.
   glGenTextures(1, &texname);
   glBindTexture(GL_TEXTURE_2D, texname);

   // Set various texture parameters
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      // TRY THIS: In above 2 lines, change "GL_NEAREST" to "GL_LINEAR".
   glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);  // Needs OpenGL 1.1
      // TRY THIS: Above, change "GL_MODULATE" to "GL_REPLACE".

   // Create the actual texture
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, img_width, img_height,
                0, GL_RGB, GL_UNSIGNED_BYTE, the_image);
}


int main(int argc, char ** argv)
{
   // Initialize OpenGL/GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

   // Make a window
   glutInitWindowSize(startwinsize, startwinsize);
   glutInitWindowPosition(50, 50);
   glutCreateWindow("CS 381 - Texture Demo");

   // Initialize GL states & register callbacks
   init();
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);
   glutSpecialFunc(special);
   glutIdleFunc(idle);

   // Do something
   glutMainLoop();

   return 0;
}
